/*
 * stdio/physfs abstraction layer 2002-12-03
 *
 * Adam D. Moss <adam@gimp.org> <aspirin@icculus.org>
 * Modified by Tree
 *
 * These wrapper macros and functions are designed to allow a program
 * to perform file I/O with identical semantics and syntax regardless
 * of whether PhysicsFS is being used or not.
 */
#ifndef __FILE_HANDLING__
#define __FILE_HANDLING__

/*
PLEASE NOTE: This license applies to abs-file.h ONLY (to make it clear that
you may embed this wrapper code within commercial software); PhysicsFS itself
is (at the time of writing) released under a different license with
additional restrictions.

Copyright (C) 2002 Adam D. Moss (the "Author").  All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is fur-
nished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CON-
NECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the Author of the
Software shall not be used in advertising or otherwise to promote the sale,
use or other dealings in this Software without prior written authorization
from the Author.
*/

#include <stdlib.h>
#include <stdio.h>

#ifdef __cplusplus
#include <vector>
#endif
/*
 * API:
 *
 * Macro/function       use like stdio equivalent...
 * --------------       ----------------------------
 * FS_FILE				FILE
 * FS_OPEN			    fopen(...)
 * FS_READ              fread(...)
 * FS_CLOSE             fclose(...)
 * FS_GETC              fgetc(...)
 * FS_GETS              fgets(...)
 * FS_EOF				feof(...)
 * FS_REWIND            rewind(...)
 *
 * FS_TELL				ftell(...)
 * FS_SEEK				fseek(...)
 * FS_FLUSH				fflush(...)
 * FS_PRINTF			fprintf(...)
 * FS_SCANF				fscanf(...)
 *
 * FS_SETBUFFER         (not a standard for stdio, does nothing there)
 */

/*
 * Important DEFINEs:
 *   It is important to define these consistantly across the various
 *   compilation modules of your program if you wish to exchange file
 *   handles between them.
 *
 *   USE_PHYSFS: Define USE_PHYSFS if PhysicsFS is being used; note that if
 *     you do intend to use PhysicsFS then you will still need to initialize
 *     PhysicsFS yourself and set up its search-paths.
 *
 * Optional DEFINEs:
 *
 *   PHYSFS_DEFAULT_READ_BUFFER <bytes>: If set then abs-file.h sets the
 *     PhysicsFS buffer size to this value whenever you open a file.  You
 *     may over-ride this on a per-filehandle basis by using the
 *     FS_SETBUFFER() macro (which simply does nothing when not using
 *     PhysicsFS).  If you have not defined this value explicitly then
 *     abs-file.h will default to the same default buffer size as used by
 *     stdio if it can be determined, or 8192 bytes otherwise.
 */
#ifndef PHYSFS_DEFAULT_READ_BUFFER
#ifdef BUFSIZ
#define PHYSFS_DEFAULT_READ_BUFFER BUFSIZ
#else
#define PHYSFS_DEFAULT_READ_BUFFER 8192
#endif
#endif

#ifdef __cplusplus

#ifdef WIN32
#define VSSCANF_MISSING
#endif

#ifdef VSSCANF_MISSING /*on a platform with vsscanf missing use this*/

static int vsscanf( const char *szString, const char* format, va_list paramList)
{
	char* szCurrentToken = (char*)strchr(format, '%');
#ifdef __cplusplus
	std::vector<void*> args; /*list of arguments*/
#else
	void* args[10];
	int count = 0;
#endif
	while (szCurrentToken != NULL)
	{
		char currentChar = *(++szCurrentToken);
		if (currentChar) /* null character = End of string */
		{
			if (currentChar == '%') /* this is just a %% to a param so skip it */
				szCurrentToken++;
			else if (currentChar != '*') /* %* causes the suppression of a parameter so ignore these */
#ifdef __cplusplus
				args.push_back(va_arg(paramList, void*));
#else
				args[count++] = va_arg(paramList, void*);
#endif
		}
		szCurrentToken = strchr(szCurrentToken,'%');
	}

	va_end(paramList);

	/* Call sscanf with correct no. of arguments...need a better way to wrap this */
#ifdef __cplusplus
	switch (args.size())
#else
	switch (count)
#endif
	{
		case  0: return 0;
		case  1: return sscanf( szString, format, args[0] );
		case  2: return sscanf( szString, format, args[0], args[1] );
		case  3: return sscanf( szString, format, args[0], args[1], args[2] );
		case  4: return sscanf( szString, format, args[0], args[1], args[2], args[3] );
		case  5: return sscanf( szString, format, args[0], args[1], args[2], args[3], args[4] );
		case  6: return sscanf( szString, format, args[0], args[1], args[2], args[3], args[4], args[5] );
		case  7: return sscanf( szString, format, args[0], args[1], args[2], args[3], args[4], args[5], args[6] );
		case  8: return sscanf( szString, format, args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7] );
		case  9: return sscanf( szString, format, args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8] );
		case 10: return sscanf( szString, format, args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8], args[9] );
	}
	/*Too many aguments should call some error handling...but for now just return -1 for failure */
	return EOF;
}
#endif /*vsscanf*/
#endif /*__cplusplus*/

#include "physfs.h"
#define FS_FILE PHYSFS_file
#define FS_SETBUFFER(fp,size) PHYSFS_setBuffer(fp,size)
#define FS_READ(p,s,n,fp) PHYSFS_read(fp,p,s,n)
#define FS_WRITE(p,s,n,fp) PHYSFS_write(fp,p,s,n)
#define FS_FLUSH(fp) PHYSFS_flush(fp)

static FS_FILE* FS_OPEN(const char *const filename, const char* type)
{
	if ((strcmp(type,"rb")==0)||(strcmp(type,"r")==0))
	{
		FS_FILE *const file = PHYSFS_openRead(filename);
#if PHYSFS_DEFAULT_READ_BUFFER
		if (file)
			FS_SETBUFFER(file, PHYSFS_DEFAULT_READ_BUFFER);
#endif
		return file;
	}
	else if ((strcmp(type,"wb")==0)||(strcmp(type,"w")==0))
	{
		/*This will be relative to the specified WriteDir*/
		return PHYSFS_openWrite(filename);
	}
	else if ((strcmp(type,"ab")==0)||(strcmp(type,"a")==0))
	{
		return PHYSFS_openAppend(filename);
	}
	/*others not supported yet*/
	return NULL;
}

static int FS_GETC(FS_FILE *const fp) 
{
  unsigned char currentCharacter;
  if (FS_READ(&currentCharacter, 1, 1, fp) != 1) 
    return EOF;

  return currentCharacter;
}

static char* FS_GETS(char* const str, const int size, FS_FILE *const fp)
{
	int index = 0;
	int currentCharacter;

	do 
	{
		if (index == size-1) 
			break;
		currentCharacter = FS_GETC(fp);
		if (currentCharacter == EOF) 
			break;
		str[index++] = (char)currentCharacter;
	} while ((currentCharacter != '\0') && (currentCharacter != -1) && (currentCharacter != '\n'));

	str[index] = '\0';
	if (index == 0) /*empty string*/
		return NULL;
	return str;
}

#define FS_CLOSE(fp) PHYSFS_close(fp)
#define FS_EOF(fp) PHYSFS_eof(fp)
#define FS_REWIND(fp) PHYSFS_seek(fp,0)
#define FS_TELL(fp) PHYSFS_tell(fp)

static int FS_SEEK(FS_FILE *fp,long index,int offset)
{
	switch (offset)
	{
		case SEEK_SET:
		{
			return PHYSFS_seek(fp,index);
		}
		case SEEK_END:
		{
			return PHYSFS_seek(fp,PHYSFS_fileLength(fp)+index);
		}
		case SEEK_CUR:
		{
			return PHYSFS_seek(fp,PHYSFS_tell(fp)+index);
		}
	}
	return -1;
}

#ifdef __cplusplus
#include <string>

static int FS_SCANF(FS_FILE *fp, const char * formatting, ...)
{
	/*/////////////////////////////////////////////////*/
	/*This is just to find the number of parameters*/
	int	numArgs = 0;
	char* currBuff = (char*)strchr(formatting, '%');
	if(currBuff == NULL) return 0;
	do
	{
		char currChar = *(++currBuff);
		if ((currChar != 0) && (currChar != '*'))
		{
			if (currChar == '%')
				currBuff++;
			else
				numArgs++;
		}
		currBuff = strchr(currBuff,'%');
	} while ( currBuff != NULL );
	/*/////////////////////////////////////////////////*/
	char lastLetter = formatting[strlen(formatting)-1];
	/*Check for simple formatting. If a format type occurs last then check for white space instead*/
	if (strchr("cdefgniouxpsXEF",lastLetter)!=NULL) lastLetter = EOF;
	/*Check for a string formatting. If % appears before [ (or no [ exists) then it is a string formatting*/
	if ((lastLetter == ']') && (strrchr(formatting,'%') < strrchr(formatting,'[')))	lastLetter = EOF;

	va_list paramList;
	va_start(paramList, formatting);

	std::string szBuffer;
	int numFound = 0;
	char letter;
	while ((letter = FS_GETC(fp))!=EOF)
	{
		szBuffer += letter;
		numFound = vsscanf(szBuffer.c_str(), formatting, paramList);
		if ((numFound == numArgs)&&((letter==EOF)||((lastLetter!=EOF)&&(letter==lastLetter))||((lastLetter==EOF)&&(isspace(letter)))))
			break;
	}

	va_end(paramList);
	return numFound;
}
/*
static int FS_PRINTF(FS_FILE *fp, const char * formatting, ...)
{
	char szEntry[1000];
	va_list paramList;
	va_start( paramList, formatting );
	int numFound = vsprintf(szEntry,formatting,paramList);
	va_end(paramList);
	FS_WRITE(szEntry,strlen(szEntry),1,fp);
	return numFound;
}
*/
#endif
#endif